;compiler: avrasm
;================================================================================================
;	32bit square root (526..542clk)
;	R16:R17 <= sqrt(R2:R3:R4:R5)
;	Author: ChaN (downloaded from ELM web (technical notes section))
;	        http://elm-chan.org/
;================================================================================================
sqrt32c:	clr	R6
	clr	R7
	clr	R8
	clr	R9
	ldi	R16,1
	ldi	R17,0
	clr	R18
	clr	R19
	ldi	R20,16
	lsl	R2
	rol	R3
	rol	R4
	rol	R5
	rol	R6
	rol	R7
	rol	R8
	rol	R9
	lsl	R2
	rol	R3
	rol	R4
	rol	R5
	rol	R6
	rol	R7
	rol	R8
	rol	R9
	brpl	PC+6
	add	R6,R16
	adc	R7,R17
	adc	R8,R18
	adc	R9,R19
	rjmp	PC+5
	sub	R6,R16
	sbc	R7,R17
	sbc	R8,R18
	sbc	R9,R19
	lsl	R16
	rol	R17
	rol	R18
	andi	R16,0b11111000
	ori	R16,0b00000101
	sbrc	R9,7
	subi	R16,2
	dec	R20
	brne	PC-34
	lsr	R18
	ror	R17
	ror	R16
	lsr	R18
	ror	R17
	ror	R16
	ret


;================================================================================================
;	y=sqrt32(x) [db], where:
;	 - output: y=R16:R17
;	 - input: x=R2:R3:R4:R5
;	using R15 as zero reg
;
;	Worst precision is some 0.6% 
;
;	Author: Stanislav Maslan, s.maslan@seznam.cz
;================================================================================================
sqrt32f:	clr	R15	;0
	tst	R5
	breq	sqrt32fn32
	;=== full 32-bit ===
	mov	R2,R4
	mov	R3,R5	;x>>=16
	rcall	sqrt32int	;s=sqrt(x)>>12
	ldi	R22,16
	mul	R20,R22
	movw	R16,R0
	mul	R21,R22
	add	R17,R0	;y=s<<12
	ret
	
sqrt32fn32:	tst	R4
	breq	sqrt32fn24
	;=== 24-bit ===
	mov	R2,R3
	mov	R3,R4	;x>>=8
	rcall	sqrt32int	;s=sqrt(x)>>8
	movw	R16,R20	;y=s<<8
	ret
	
sqrt32fn24:	tst	R3
	breq	sqrt32fn16
	;=== 16-bit ===
	rcall	sqrt32int	;s=sqrt(x)>>4
	ldi	R22,16
	mul	R21,R22
	movw	R16,R0
	mul	R20,R22
	add	R16,R1
	adc	R17,R15	;y=s<<4
	ret
	;=== 8-bit ===
sqrt32fn16:	ldi	R16,8
	cp	R2,R16
	brsh	sqrt32fn16lp
	;x<8 - switch to higher precision table
	ldi	R16,16*2
	mul	R2,R16
	movw	R30,R0
	subiw	R30,-(TSqrt256p1*2+1)
	rjmp	sqrt32fn16hp
	;x>8 - standard precision table
sqrt32fn16lp:	ldi	R16,8
	sub	R2,R16
	ldi	R16,2
	mul	R16,R2
	movw	R30,R0
	subiw	R30,-(TSqrt256p2*2+1)
sqrt32fn16hp:	lpm	R16,Z
	clr	R17	;y=sqrt(x)
	ret

;------------------------------------------------------------------------------------------------
;	return y=sqrt(x) from lookup table where: 
;	int part: x=R3
;	fraction: f=R2 [fx 0.8]
;	output: y=R20:R21 [fx 8.8]
;	using R15 as zero reg 
;------------------------------------------------------------------------------------------------
sqrt32int:	ldi	R16,8
	cp	R3,R16
	brsh	sqrt32intLP
	;x<8 - use higher precision lookup
	ldi	R22,16
	clr	R18
	mul	R2,R22
	movw	R16,R0
	mul	R3,R22
	addw	R17,R0	;(f:x)<<4
	mov	R2,R16	;f=((f:x)<<4)&0x000FF
	mov	R30,R17
	mov	R31,R18
	lslw	R30	;x=((f:x)>>4)&0x0FFFF
	subiw	R30,-TSqrt256p1*2	;Z=TSqrt256p1[x]	
	rjmp	sqrt32intHP
	;x>=8 - use standard precision loopup 
sqrt32intLP:	ldi	R16,8
	sub	R3,R16
	ldi	R16,2
	mul	R3,R16
	movw	R30,R0
	subiw	R30,-TSqrt256p2*2	;Z=TSqrt256p2[x]
sqrt32intHP:	lpmw	R20,Z+	;y1=sqrt(x)=*Z++ [fx 8.8]
	lpmw	R22,Z+	;y2=sqrt(x+1)=*Z++ [fx 8.8]
	subw	R22,R20	;dy=y2-y1 [fx 8.8]
	
	mul	R23,R2	
	addw	R20,R0
	mul	R22,R2
	add	R20,R1
	adc	R21,R15	;y=sqrt(y1+f*dy) [fx 8.8]
	ret	

	;sqrt(x) [fx 8.8], for x=0,(1/16),(2/16)...(127/16),8 
TSqrt256p1:	.dw	0,64,91,111,128,143,157,169
	.dw	181,192,202,212,222,231,239,248
	.dw	256,264,272,279,286,293,300,307
	.dw	314,320,326,333,339,345,351,356
	.dw	362,368,373,379,384,389,395,400
	.dw	405,410,415,420,425,429,434,439
	.dw	443,448,453,457,462,466,470,475
	.dw	479,483,487,492,496,500,504,508
	.dw	512,516,520,524,528,532,535,539
	.dw	543,547,551,554,558,562,565,569
	.dw	572,576,580,583,587,590,594,597
	.dw	600,604,607,611,614,617,621,624
	.dw	627,630,634,637,640,643,646,650
	.dw	653,656,659,662,665,668,671,674
	.dw	677,680,683,686,689,692,695,698
	.dw	701,704,707,710,713,716,718,721
	.dw	724
	;sqrt(x) [fx 8.8], for x=8...256
TSqrt256p2:	.dw	724,768,810,849,887,923,958,991
	.dw	1024,1056,1086,1116,1145,1173,1201,1228
	.dw	1254,1280,1305,1330,1355,1379,1402,1425
	.dw	1448,1471,1493,1515,1536,1557,1578,1599
	.dw	1619,1639,1659,1679,1698,1717,1736,1755
	.dw	1774,1792,1810,1828,1846,1864,1881,1899
	.dw	1916,1933,1950,1966,1983,1999,2016,2032
	.dw	2048,2064,2080,2095,2111,2126,2142,2157
	.dw	2172,2187,2202,2217,2232,2246,2261,2275
	.dw	2290,2304,2318,2332,2346,2360,2374,2388
	.dw	2401,2415,2429,2442,2455,2469,2482,2495
	.dw	2508,2521,2534,2547,2560,2573,2585,2598
	.dw	2611,2623,2636,2648,2660,2673,2685,2697
	.dw	2709,2721,2733,2745,2757,2769,2781,2793
	.dw	2804,2816,2828,2839,2851,2862,2874,2885
	.dw	2896,2908,2919,2930,2941,2952,2963,2974
	.dw	2985,2996,3007,3018,3029,3040,3051,3061
	.dw	3072,3083,3093,3104,3114,3125,3135,3146
	.dw	3156,3167,3177,3187,3197,3208,3218,3228
	.dw	3238,3248,3258,3268,3278,3288,3298,3308
	.dw	3318,3328,3338,3348,3357,3367,3377,3387
	.dw	3396,3406,3415,3425,3435,3444,3454,3463
	.dw	3473,3482,3491,3501,3510,3519,3529,3538
	.dw	3547,3556,3566,3575,3584,3593,3602,3611
	.dw	3620,3629,3638,3647,3656,3665,3674,3683
	.dw	3692,3701,3710,3719,3727,3736,3745,3754
	.dw	3762,3771,3780,3788,3797,3806,3814,3823
	.dw	3831,3840,3849,3857,3866,3874,3882,3891
	.dw	3899,3908,3916,3924,3933,3941,3949,3958
	.dw	3966,3974,3982,3991,3999,4007,4015,4023
	.dw	4031,4040,4048,4056,4064,4072,4080,4088
	.dw	4096
	
	




;================================================================================================
;	y=20*log10(x/log10rm) [db], where:
;	 - output: y=signed(R16:R17) [fx 8.8]
;	 - input: x=unsinged(R16:R17)
;	using R15 as zero reg
;
;	Precision is usualy +/-1, max errors are at lowest values of lookups where
;	log function is stronly nonlinear (error<0.2% except lowest amplitudes)   
;
;	examples:
;	ldiw	R16,16383
;	rcall	log10f	;R16:R17=0 (0/256=0dB)
;	ldiw	R16,164
;	rcall	log10f	;R16:R17=-10237 (-10237/256=-39.99dB)
;
;	Author: Stanislav Maslan, s.maslan@seznam.cz
;================================================================================================
	;theese values are linked to lookups TLog10p1,TLog10p2,TLog10p3
	;!!! can't be changed without recalculation of lookups !!!
	.equ	log10rm=16383	;x value for y=0dB
	.equ	log10rs0=2	;table 0 step
	.equ	log10rr0=64	;table 0 max
	.equ	log10rs1=16	;table 1 step
	.equ	log10rr1=1280	;table 1 max
	.equ	log10rr2=16383	;table 2 max
	
log10f:	clr	R15
	cpiwh	R16,R18,log10rr0
	brsh	log10fp1N
	;=== TLog10p1 ===
	ldi	R18,256/log10rs0
	mul	R16,R18
	movw	R16,R0
	ldiw	R30,TLog10p1*2
	rjmp	log10fCM
		
log10fp1N:	cpiwh	R16,R18,log10rr1
	brsh	log10fp2N
	;=== TLog10p2 ===
	subiw	R16,log10rr0
	ldi	R20,256/log10rs1
	mul	R16,R20
	movw	R18,R0
	mul	R17,R20
	add	R19,R0
	movw	R16,R18
	ldiw	R30,TLog10p2*2
	rjmp	log10fCM
	
	;=== TLog10p3 ===
log10fp2N:	cpiwh	R16,R18,log10rr2
	brlo	PC+1+2
	ldiw	R16,log10rr2
	subiw	R16,log10rr1
	ldiw	R30,TLog10p3*2
	rjmp	log10fCM
		
		

;------------------------------------------------------------------------------------------------
;	y=20*log10(z) [db], where:
;	 - output: y=R16:R17 [fx 8.8]
;	 - input: z=R16:R17 [fx 8.8]
;	    - integer: x=R17
;	    - fraction: f=R16 [0.8]
;	using R15 as zero reg
;------------------------------------------------------------------------------------------------
log10fCM:	ldi	R18,2
	mul	R17,R18
	addw	R30,R0	;lookup pointer
	;get values
	lpmw	R18,Z+	;lg1=log10(x)
	lpmw	R20,Z+	;lg2=log10(x+1)
	subw	R20,R18	;ld=lg2-lg1
	;linear interpolation
	mulsu	R21,R16
	addw	R18,R0
	mul	R20,R16
	lsl	R0	;round
	adc	R18,R1
	adc	R19,R15	;lg=lg1-ld*f
	;return result
	movw	R16,R18
	ret

	

	;y=20*log10(x/16383) [dB] [fx 8.8], for x=0...32 [fx 15.1]
	;y range: -84.29 to -48.16dB
TLog10p1:	.dw	-23118,-20036,-18495,-17594,-16954,-16458,-16052,-15710
	.dw	-15413,-15151,-14916,-14704,-14511,-14333,-14168,-14015
	.dw	-13871,-13737,-13609,-13489,-13375,-13267,-13163,-13064
	.dw	-12970,-12879,-12792,-12708,-12627,-12549,-12474,-12401
	.dw	-12330

	;y=20*log10(x/16383) [dB] [fx 8.8], for x=4...80 [fx 12.4]
	;y range: -48.16 to -22.14dB
TLog10p2:	.dw	-12330,-11834,-11428,-11086,-10789,-10527,-10293,-10081
	.dw	 -9887, -9709, -9544, -9391, -9248, -9113, -8986, -8865
	.dw	 -8751, -8643, -8539, -8441, -8346, -8255, -8168, -8084
	.dw	 -8003, -7925, -7850, -7777, -7706, -7638, -7571, -7507
	.dw	 -7444, -7383, -7324, -7266, -7210, -7155, -7102, -7049
	.dw	 -6998, -6948, -6899, -6851, -6805, -6759, -6714, -6670
	.dw	 -6627, -6584, -6543, -6502, -6462, -6423, -6384, -6346
	.dw	 -6308, -6272, -6236, -6200, -6165, -6130, -6097, -6063
	.dw	 -6030, -5998, -5966, -5934, -5903, -5872, -5842, -5812
	.dw	 -5783, -5754, -5725, -5697, -5669

	;y=20*log10(x/16383) [dB] [fx 8.8], for x=5...63 [fx 8.8]
	;y range: -22.14 to 0.00dB
TLog10p3:	.dw	 -5669, -5263, -4921, -4624, -4362, -4128, -3916, -3722
	.dw	 -3544, -3379, -3226, -3082, -2948, -2821, -2700, -2586
	.dw	 -2478, -2374, -2275, -2181, -2090, -2003, -1919, -1838
	.dw	 -1760, -1685, -1612, -1541, -1473, -1406, -1342, -1279
	.dw	 -1218, -1159, -1101, -1045,  -990,  -936,  -884,  -833
	.dw	  -783,  -734,  -686,  -640,  -594,  -549,  -505,  -462
	.dw	  -419,  -378,  -337,  -297,  -257,  -219,  -181,  -143
	.dw	  -107,   -70,   -35,     0	

	